@@ -25,7 +25,6 @@
 #include "pango-glyph-item.h"
 #include "pango-layout-private.h"
 #include "pango-font-private.h"
-#include "pango-attributes-private.h"
 #include "pango-impl-utils.h"
 
 typedef struct _EllipsizeState EllipsizeState;
@@ -106,29 +105,21 @@ struct _EllipsizeState
 
   LineIter gap_end_iter;	/* Iterator pointing to last cluster in gap */
   int gap_end_x;		/* x position of end of gap, in Pango units */
-
-  PangoShapeFlags shape_flags;
 };
 
 /* Compute global information needed for the itemization process
  */
 static void
 init_state (EllipsizeState  *state,
-            PangoLayoutLine *line,
-            PangoAttrList   *attrs,
-            PangoShapeFlags  shape_flags)
+	    PangoLayoutLine *line,
+	    PangoAttrList   *attrs)
 {
   GSList *l;
   int i;
   int start_offset;
 
   state->layout = line->layout;
-  if (attrs)
-    state->attrs = pango_attr_list_ref (attrs);
-  else
-    state->attrs = pango_attr_list_new ();
-
-  state->shape_flags = shape_flags;
+  state->attrs = attrs;
 
   state->n_runs = g_slist_length (line->runs);
   state->run_info = g_new (RunInfo, state->n_runs);
@@ -160,7 +151,6 @@ init_state (EllipsizeState  *state,
 static void
 free_state (EllipsizeState *state)
 {
-  pango_attr_list_unref (state->attrs);
   if (state->line_start_attr)
     pango_attr_iterator_destroy (state->line_start_attr);
   if (state->gap_start_attr)
@@ -300,18 +290,15 @@ itemize_text (EllipsizeState *state,
 static void
 shape_ellipsis (EllipsizeState *state)
 {
-  PangoAttrList attrs;
+  PangoAttrList *attrs = pango_attr_list_new ();
   GSList *run_attrs;
   PangoItem *item;
   PangoGlyphString *glyphs;
   GSList *l;
   PangoAttribute *fallback;
   const char *ellipsis_text;
-  int len;
   int i;
 
-  _pango_attr_list_init (&attrs);
-
   /* Create/reset state->ellipsis_run
    */
   if (!state->ellipsis_run)
@@ -336,7 +323,7 @@ shape_ellipsis (EllipsizeState *state)
       attr->start_index = 0;
       attr->end_index = G_MAXINT;
 
-      pango_attr_list_insert (&attrs, attr);
+      pango_attr_list_insert (attrs, attr);
     }
 
   g_slist_free (run_attrs);
@@ -344,7 +331,7 @@ shape_ellipsis (EllipsizeState *state)
   fallback = pango_attr_fallback_new (FALSE);
   fallback->start_index = 0;
   fallback->end_index = G_MAXINT;
-  pango_attr_list_insert (&attrs, fallback);
+  pango_attr_list_insert (attrs, fallback);
 
   /* First try using a specific ellipsis character in the best matching font
    */
@@ -353,7 +340,7 @@ shape_ellipsis (EllipsizeState *state)
   else
     ellipsis_text = "\342\200\246";	/* U+2026: HORIZONTAL ELLIPSIS */
 
-  item = itemize_text (state, ellipsis_text, &attrs);
+  item = itemize_text (state, ellipsis_text, attrs);
 
   /* If that fails we use "..." in the first matching font
    */
@@ -368,10 +355,10 @@ shape_ellipsis (EllipsizeState *state)
       ((PangoAttrInt *)fallback)->value = TRUE;
 
       ellipsis_text = "...";
-      item = itemize_text (state, ellipsis_text, &attrs);
+      item = itemize_text (state, ellipsis_text, attrs);
     }
 
-  _pango_attr_list_destroy (&attrs);
+  pango_attr_list_unref (attrs);
 
   state->ellipsis_run->item = item;
 
@@ -379,11 +366,8 @@ shape_ellipsis (EllipsizeState *state)
    */
   glyphs = state->ellipsis_run->glyphs;
 
-  len = strlen (ellipsis_text);
-  pango_shape_with_flags (ellipsis_text, len,
-                          ellipsis_text, len,
-	                  &item->analysis, glyphs,
-                          state->shape_flags);
+  pango_shape (ellipsis_text, strlen (ellipsis_text),
+	       &item->analysis, glyphs);
 
   state->ellipsis_width = 0;
   for (i = 0; i < glyphs->num_glyphs; i++)
@@ -734,7 +718,6 @@ current_width (EllipsizeState *state)
  * _pango_layout_line_ellipsize:
  * @line: a #PangoLayoutLine
  * @attrs: Attributes being used for itemization/shaping
- * @shape_flags: Flags to use when shaping
  *
  * Given a #PangoLayoutLine with the runs still in logical order, ellipsize
  * it according the layout's policy to fit within the set width of the layout.
@@ -744,7 +727,6 @@ current_width (EllipsizeState *state)
 gboolean
 _pango_layout_line_ellipsize (PangoLayoutLine *line,
 			      PangoAttrList   *attrs,
-                              PangoShapeFlags  shape_flags,
 			      int              goal_width)
 {
   EllipsizeState state;
@@ -752,7 +734,7 @@ _pango_layout_line_ellipsize (PangoLayoutLine *line,
 
   g_return_val_if_fail (line->layout->ellipsize != PANGO_ELLIPSIZE_NONE && goal_width >= 0, is_ellipsized);
 
-  init_state (&state, line, attrs, shape_flags);
+  init_state (&state, line, attrs);
 
   if (state.total_width <= goal_width)
     goto out;
