@@ -29,7 +29,7 @@
  * implementation of fonts. The #PangoFont structure
  * represents an abstract rendering-system-independent font.
  * Pango provides routines to list available fonts, and
- * to load a font of a given description.
+ * to load a font matching a given description.
  */
 
 #include "config.h"
@@ -38,7 +38,7 @@
 #include <string.h>
 
 #include "pango-types.h"
-#include "pango-font.h"
+#include "pango-font-private.h"
 #include "pango-fontmap.h"
 #include "pango-impl-utils.h"
 
@@ -1527,7 +1527,7 @@ parse_field (const char *what,
 /**
  * pango_parse_style:
  * @str: a string to parse.
- * @style: (out caller-allocates): a #PangoStyle to store the result
+ * @style: (out): a #PangoStyle to store the result
  *   in.
  * @warn: if %TRUE, issue a g_warning() on bad input.
  *
@@ -1548,7 +1548,7 @@ pango_parse_style (const char *str,
 /**
  * pango_parse_variant:
  * @str: a string to parse.
- * @variant: (out caller-allocates): a #PangoVariant to store the
+ * @variant: (out): a #PangoVariant to store the
  *   result in.
  * @warn: if %TRUE, issue a g_warning() on bad input.
  *
@@ -1569,7 +1569,7 @@ pango_parse_variant (const char   *str,
 /**
  * pango_parse_weight:
  * @str: a string to parse.
- * @weight: (out caller-allocates): a #PangoWeight to store the result
+ * @weight: (out): a #PangoWeight to store the result
  *   in.
  * @warn: if %TRUE, issue a g_warning() on bad input.
  *
@@ -1590,7 +1590,7 @@ pango_parse_weight (const char  *str,
 /**
  * pango_parse_stretch:
  * @str: a string to parse.
- * @stretch: (out caller-allocates): a #PangoStretch to store the
+ * @stretch: (out): a #PangoStretch to store the
  *   result in.
  * @warn: if %TRUE, issue a g_warning() on bad input.
  *
@@ -1617,11 +1617,29 @@ pango_parse_stretch (const char   *str,
  * PangoFont
  */
 
-G_DEFINE_ABSTRACT_TYPE (PangoFont, pango_font, G_TYPE_OBJECT)
+typedef struct {
+  hb_font_t *hb_font;
+} PangoFontPrivate;
+
+G_DEFINE_ABSTRACT_TYPE_WITH_PRIVATE (PangoFont, pango_font, G_TYPE_OBJECT)
+
+static void
+pango_font_finalize (GObject *object)
+{
+  PangoFont *font = PANGO_FONT (object);
+  PangoFontPrivate *priv = pango_font_get_instance_private (font);
+
+  hb_font_destroy (priv->hb_font);
+
+  G_OBJECT_CLASS (pango_font_parent_class)->finalize (object);
+}
 
 static void
 pango_font_class_init (PangoFontClass *class G_GNUC_UNUSED)
 {
+  GObjectClass *object_class = G_OBJECT_CLASS (class);
+
+  object_class->finalize = pango_font_finalize;
 }
 
 static void
@@ -1702,20 +1720,14 @@ pango_font_get_coverage (PangoFont     *font,
  * language tag and character point.
  *
  * Return value: (transfer none): the best matching shaper.
+ * Deprecated: Shape engines are no longer used
  **/
 PangoEngineShape *
 pango_font_find_shaper (PangoFont     *font,
 			PangoLanguage *language,
 			guint32        ch)
 {
-  PangoEngineShape* shaper;
-
-  if (G_UNLIKELY (!font))
-    return NULL;
-
-  shaper = PANGO_FONT_GET_CLASS (font)->find_shaper (font, language, ch);
-
-  return shaper;
+  return NULL;
 }
 
 /**
@@ -1793,6 +1805,7 @@ pango_font_get_metrics (PangoFont        *font,
 
       metrics->ascent = PANGO_SCALE * PANGO_UNKNOWN_GLYPH_HEIGHT;
       metrics->descent = 0;
+      metrics->height = 0;
       metrics->approximate_char_width = PANGO_SCALE * PANGO_UNKNOWN_GLYPH_WIDTH;
       metrics->approximate_digit_width = PANGO_SCALE * PANGO_UNKNOWN_GLYPH_WIDTH;
       metrics->underline_position = -PANGO_SCALE;
@@ -1837,6 +1850,38 @@ pango_font_get_font_map (PangoFont *font)
     return NULL;
 }
 
+/**
+ * pango_font_get_hb_font:
+ * @font: a #PangoFont
+ *
+ * Get a hb_font_t object backing this font.
+ *
+ * Note that the objects returned by this function
+ * are cached and immutable. If you need to make
+ * changes to the hb_font_t, use hb_font_create_sub_font().
+ *
+ * Returns: (transfer none) (nullable): the hb_font_t object backing the
+ *          font, or %NULL if the font does not have one
+ *
+ * Since: 1.44
+ */
+hb_font_t *
+pango_font_get_hb_font (PangoFont *font)
+{
+  PangoFontPrivate *priv = pango_font_get_instance_private (font);
+
+  g_return_val_if_fail (PANGO_IS_FONT (font), NULL);
+
+  if (priv->hb_font)
+    return priv->hb_font;
+
+  priv->hb_font = PANGO_FONT_GET_CLASS (font)->create_hb_font (font);
+
+  hb_font_make_immutable (priv->hb_font);
+
+  return priv->hb_font;
+}
+
 G_DEFINE_BOXED_TYPE (PangoFontMetrics, pango_font_metrics,
                      pango_font_metrics_ref,
                      pango_font_metrics_unref);
@@ -1939,6 +1984,28 @@ pango_font_metrics_get_descent (PangoFontMetrics *metrics)
   return metrics->descent;
 }
 
+/**
+ * pango_font_metrics_get_height:
+ * @metrics: a #PangoFontMetrics structure
+ *
+ * Gets the line height from a font metrics structure. The
+ * line height is the distance between successive baselines
+ * in wrapped text.
+ *
+ * If the line height is not available, 0 is returned.
+ *
+ * Return value: the height, in Pango units
+ *
+ * Since: 1.44
+ */
+int
+pango_font_metrics_get_height (PangoFontMetrics *metrics)
+{
+  g_return_val_if_fail (metrics != NULL, 0);
+
+  return metrics->height;
+}
+
 /**
  * pango_font_metrics_get_approximate_char_width:
  * @metrics: a #PangoFontMetrics structure
@@ -2283,3 +2350,50 @@ pango_font_face_list_sizes (PangoFontFace  *face,
       *n_sizes = 0;
     }
 }
+
+/**
+ * pango_font_has_char:
+ * @font: a #PangoFont
+ * @wc: a Unicode character
+ *
+ * Returns whether the font provides a glyph for this character.
+ *
+ * Returns %TRUE if @font can render @wc
+ *
+ * Since: 1.44
+ */
+gboolean
+pango_font_has_char (PangoFont *font,
+                     gunichar   wc)
+{
+  PangoCoverage *coverage = pango_font_get_coverage (font, pango_language_get_default ());
+  PangoCoverageLevel result = pango_coverage_get (coverage, wc);
+  pango_coverage_unref (coverage);
+  return result != PANGO_COVERAGE_NONE;
+}
+
+/**
+ * pango_font_get_features:
+ * @font: a #PangoFont
+ * @features: (out caller-allocates) (array length=len): Array to features in
+ * @len: the length of @features
+ * @num_features: (inout): the number of used items in @features
+ *
+ * Obtain the OpenType features that are provided by the font.
+ * These are passed to the rendering system, together with features
+ * that have been explicitly set via attributes.
+ *
+ * Note that this does not include OpenType features which the
+ * rendering system enables by default.
+ *
+ * Since: 1.44
+ */
+void
+pango_font_get_features (PangoFont    *font,
+                         hb_feature_t *features,
+                         guint         len,
+                         guint        *num_features)
+{
+  if (PANGO_FONT_GET_CLASS (font)->get_features)
+    PANGO_FONT_GET_CLASS (font)->get_features (font, features, len, num_features);
+}
