@@ -22,7 +22,7 @@
 /**
  * SECTION:text-attributes
  * @short_description:Font and other attributes for annotating text
- * @title:Text Attributes
+ * @title:Attributes
  *
  * Attributed text is used in a number of places in Pango. It
  * is used as the input to the itemization process and also when
@@ -192,6 +192,10 @@ pango_attribute_destroy (PangoAttribute *attr)
   attr->klass->destroy (attr);
 }
 
+G_DEFINE_BOXED_TYPE (PangoAttribute, pango_attribute,
+                     pango_attribute_copy,
+                     pango_attribute_destroy);
+
 /**
  * pango_attribute_equal:
  * @attr1: a #PangoAttribute
@@ -552,7 +556,7 @@ pango_attr_size_new_internal (int size,
 
 /**
  * pango_attr_size_new:
- * @size: the font size, in %PANGO_SCALE<!-- -->ths of a point.
+ * @size: the font size, in %PANGO_SCALEths of a point.
  *
  * Create a new font-size attribute in fractional points.
  *
@@ -567,7 +571,7 @@ pango_attr_size_new (int size)
 
 /**
  * pango_attr_size_new_absolute:
- * @size: the font size, in %PANGO_SCALE<!-- -->ths of a device unit.
+ * @size: the font size, in %PANGO_SCALEths of a device unit.
  *
  * Create a new font-size attribute in device units.
  *
@@ -1175,6 +1179,58 @@ pango_attr_background_alpha_new (guint16 alpha)
   return pango_attr_int_new (&klass, (int)alpha);
 }
 
+/**
+ * pango_attr_allow_breaks_new:
+ * @allow_breaks: %TRUE if we line breaks are allowed
+ *
+ * Create a new allow-breaks attribute.
+ *
+ * If breaks are disabled, the range will be kept in a
+ * single run, as far as possible.
+ *
+ * Return value: (transfer full): the newly allocated #PangoAttribute,
+ *               which should be freed with pango_attribute_destroy()
+ *
+ * Since: 1.44
+ */
+PangoAttribute *
+pango_attr_allow_breaks_new (gboolean allow_breaks)
+{
+  static const PangoAttrClass klass = {
+    PANGO_ATTR_ALLOW_BREAKS,
+    pango_attr_int_copy,
+    pango_attr_int_destroy,
+    pango_attr_int_equal,
+  };
+
+  return pango_attr_int_new (&klass, (int)allow_breaks);
+}
+
+/**
+ * pango_attr_show_new:
+ * @show: #PangoShowFlags to apply
+ *
+ * Create a new attribute that influences how invisible
+ * characters are rendered.
+ *
+ * Return value: (transfer full): the newly allocated #PangoAttribute,
+ *               which should be freed with pango_attribute_destroy().
+ *
+ * Since: 1.44
+ **/
+PangoAttribute *
+pango_attr_show_new (PangoShowFlags flags)
+{
+  static const PangoAttrClass klass = {
+    PANGO_ATTR_SHOW,
+    pango_attr_int_copy,
+    pango_attr_int_destroy,
+    pango_attr_int_equal,
+  };
+
+  return pango_attr_int_new (&klass, (int)flags);
+}
+
 /*
  * Attribute List
  */
@@ -1339,9 +1395,6 @@ pango_attr_list_insert_internal (PangoAttrList  *list,
 	      else
 		list->attributes = link;
 
-	      if (!tmp_list)
-		list->attributes_tail = link;
-
 	      break;
 	    }
 
@@ -1597,6 +1650,87 @@ pango_attr_list_change (PangoAttrList  *list,
     }
 }
 
+/**
+ * pango_attr_list_update:
+ * @list: a #PangoAttrList
+ * @pos: the position of the change
+ * @remove: the number of removed bytes
+ * @add: the number of added bytes
+ *
+ * Update indices of attributes in @list for
+ * a change in the text they refer to.
+ *
+ * The change that this function applies is
+ * removing @remove bytes at position @pos
+ * and inserting @add bytes instead.
+ *
+ * Attributes that fall entirely in the
+ * (@pos, @pos + @remove) range are removed.
+ *
+ * Attributes that start or end inside the
+ * (@pos, @pos + @remove) range are shortened to
+ * reflect the removal.
+ *
+ * Attributes start and end positions are updated
+ * if they are behind @pos + @remove.
+ *
+ * Since: 1.44
+ */
+void
+pango_attr_list_update (PangoAttrList *list,
+                        int             pos,
+                        int             remove,
+                        int             add)
+{
+  GSList *l, *prev, *next;
+
+   prev = NULL;
+   l = list->attributes;
+   while (l)
+    {
+      PangoAttribute *attr = l->data;
+      next = l->next;
+
+      if (attr->start_index >= pos &&
+          attr->end_index < pos + remove)
+        {
+          pango_attribute_destroy (attr);
+          if (prev == NULL)
+            list->attributes = next;
+          else
+            prev->next = next;
+
+          g_slist_free_1 (l);
+        }
+      else
+        {
+          prev = l;
+
+          if (attr->start_index >= pos &&
+              attr->start_index < pos + remove)
+            {
+              attr->start_index = pos + add;
+            }
+          else if (attr->start_index >= pos + remove)
+            {
+              attr->start_index += add - remove;
+            }
+
+          if (attr->end_index >= pos &&
+              attr->end_index < pos + remove)
+            {
+              attr->end_index = pos;
+            }
+          else if (attr->end_index >= pos + remove)
+            {
+              attr->end_index += add - remove;
+            }
+        }
+
+      l = next;
+    }
+}
+
 /**
  * pango_attr_list_splice:
  * @list: a #PangoAttrList
@@ -1680,6 +1814,27 @@ pango_attr_list_splice (PangoAttrList *list,
 #undef CLAMP_ADD
 }
 
+/**
+ * pango_attr_list_get_attributes:
+ * @list: a #PangoAttrList
+ *
+ * Gets a list of all attributes in @list.
+ *
+ * Return value: (element-type Pango.Attribute) (transfer full):
+ *   a list of all attributes in @list. To free this value, call
+ *   pango_attribute_destroy() on each value and g_slist_free()
+ *   on the list.
+ *
+ * Since: 1.44
+ */
+GSList *
+pango_attr_list_get_attributes (PangoAttrList *list)
+{
+  g_return_val_if_fail (list != NULL, NULL);
+
+  return g_slist_copy_deep (list->attributes, (GCopyFunc)pango_attribute_copy, NULL);
+}
+
 /**
  * pango_attr_list_get_iterator:
  * @list: a #PangoAttrList
