@@ -28,6 +28,8 @@
  * Unicode characters and converts it into glyphs.
  * The functions described in this section accomplish
  * various steps of this process.
+ *
+ * ![](pipeline.png)
  */
 #include "config.h"
 #include <string.h>
@@ -36,10 +38,23 @@
 #include "pango-context.h"
 #include "pango-impl-utils.h"
 
-#include "pango-engine-private.h"
+#include "pango-font-private.h"
+#include "pango-fontset-private.h"
+#include "pango-fontmap-private.h"
 #include "pango-script-private.h"
 #include "pango-emoji-private.h"
 
+/**
+ * SECTION:context
+ * @title:Contexts
+ * @short_description: Global context object
+ *
+ * The #PangoContext structure stores global information
+ * influencing Pango's operation, such as the fontmap used
+ * to look up fonts, and default values such as the default
+ * language, default gravity, or default font.
+ */
+
 /**
  * PangoContext:
  *
@@ -601,7 +616,7 @@ advance_attr_iterator_to (PangoAttrIterator *iterator,
 }
 
 /***************************************************************************
- * We cache the results of character,fontset => font,shaper in a hash table
+ * We cache the results of character,fontset => font in a hash table
  ***************************************************************************/
 
 typedef struct {
@@ -634,7 +649,7 @@ get_font_cache (PangoFontset *fontset)
 
   static GQuark cache_quark = 0; /* MT-safe */
   if (G_UNLIKELY (!cache_quark))
-    cache_quark = g_quark_from_static_string ("pango-shaper-font-cache");
+    cache_quark = g_quark_from_static_string ("pango-font-cache");
 
 retry:
   cache = g_object_get_qdata (G_OBJECT (fontset), cache_quark);
@@ -706,7 +721,7 @@ struct _PangoWidthIter
 	const gchar *text_end;
 	const gchar *start;
 	const gchar *end;
-	gboolean wide;
+	gboolean     upright;
 };
 
 typedef struct _ItemizeState ItemizeState;
@@ -755,7 +770,6 @@ struct _ItemizeState
   PangoEmojiIter emoji_iter;
 
   PangoLanguage *derived_lang;
-  PangoEngineLang *lang_engine;
 
   PangoFontset *current_fonts;
   FontCache *cache;
@@ -853,25 +867,58 @@ update_end (ItemizeState *state)
     state->run_end = state->emoji_iter.end;
 }
 
-/* g_unichar_iswide() uses EastAsianWidth, which is broken.
- * We should switch to using VerticalTextLayout:
- * http://www.unicode.org/reports/tr50/#Data50
- *
- * In the mean time, fixup Hangul jamo to be all wide so we
- * don't break run in the middle.  The EastAsianWidth has
- * 'W' for L-jamo, and 'N' for T and V jamo!
- *
- * https://bugzilla.gnome.org/show_bug.cgi?id=705727
- */
 static gboolean
-width_iter_iswide (gunichar ch)
+width_iter_is_upright (gunichar ch)
 {
-  if ((0x1100u <= ch && ch <= 0x11FFu) ||
-      (0xA960u <= ch && ch <= 0xA97Cu) ||
-      (0xD7B0u <= ch && ch <= 0xD7FBu))
-    return TRUE;
+  /* https://www.unicode.org/Public/11.0.0/ucd/VerticalOrientation.txt
+   * VO=U or Tu table generated by tools/gen-vertical-orientation-U-table.py.
+   *
+   * FIXME: In the future, If GLib supports VerticalOrientation, please use it.
+   */
+  static const gunichar upright[][2] = {
+    {0x00A7, 0x00A7}, {0x00A9, 0x00A9}, {0x00AE, 0x00AE}, {0x00B1, 0x00B1},
+    {0x00BC, 0x00BE}, {0x00D7, 0x00D7}, {0x00F7, 0x00F7}, {0x02EA, 0x02EB},
+    {0x1100, 0x11FF}, {0x1401, 0x167F}, {0x18B0, 0x18FF}, {0x2016, 0x2016},
+    {0x2020, 0x2021}, {0x2030, 0x2031}, {0x203B, 0x203C}, {0x2042, 0x2042},
+    {0x2047, 0x2049}, {0x2051, 0x2051}, {0x2065, 0x2065}, {0x20DD, 0x20E0},
+    {0x20E2, 0x20E4}, {0x2100, 0x2101}, {0x2103, 0x2109}, {0x210F, 0x210F},
+    {0x2113, 0x2114}, {0x2116, 0x2117}, {0x211E, 0x2123}, {0x2125, 0x2125},
+    {0x2127, 0x2127}, {0x2129, 0x2129}, {0x212E, 0x212E}, {0x2135, 0x213F},
+    {0x2145, 0x214A}, {0x214C, 0x214D}, {0x214F, 0x2189}, {0x218C, 0x218F},
+    {0x221E, 0x221E}, {0x2234, 0x2235}, {0x2300, 0x2307}, {0x230C, 0x231F},
+    {0x2324, 0x2328}, {0x232B, 0x232B}, {0x237D, 0x239A}, {0x23BE, 0x23CD},
+    {0x23CF, 0x23CF}, {0x23D1, 0x23DB}, {0x23E2, 0x2422}, {0x2424, 0x24FF},
+    {0x25A0, 0x2619}, {0x2620, 0x2767}, {0x2776, 0x2793}, {0x2B12, 0x2B2F},
+    {0x2B50, 0x2B59}, {0x2BB8, 0x2BD1}, {0x2BD3, 0x2BEB}, {0x2BF0, 0x2BFF},
+    {0x2E80, 0x3007}, {0x3012, 0x3013}, {0x3020, 0x302F}, {0x3031, 0x309F},
+    {0x30A1, 0x30FB}, {0x30FD, 0xA4CF}, {0xA960, 0xA97F}, {0xAC00, 0xD7FF},
+    {0xE000, 0xFAFF}, {0xFE10, 0xFE1F}, {0xFE30, 0xFE48}, {0xFE50, 0xFE57},
+    {0xFE5F, 0xFE62}, {0xFE67, 0xFE6F}, {0xFF01, 0xFF07}, {0xFF0A, 0xFF0C},
+    {0xFF0E, 0xFF19}, {0xFF1F, 0xFF3A}, {0xFF3C, 0xFF3C}, {0xFF3E, 0xFF3E},
+    {0xFF40, 0xFF5A}, {0xFFE0, 0xFFE2}, {0xFFE4, 0xFFE7}, {0xFFF0, 0xFFF8},
+    {0xFFFC, 0xFFFD}, {0x10980, 0x1099F}, {0x11580, 0x115FF}, {0x11A00, 0x11AAF},
+    {0x13000, 0x1342F}, {0x14400, 0x1467F}, {0x16FE0, 0x18AFF}, {0x1B000, 0x1B12F},
+    {0x1B170, 0x1B2FF}, {0x1D000, 0x1D1FF}, {0x1D2E0, 0x1D37F}, {0x1D800, 0x1DAAF},
+    {0x1F000, 0x1F7FF}, {0x1F900, 0x1FA6F}, {0x20000, 0x2FFFD}, {0x30000, 0x3FFFD},
+    {0xF0000, 0xFFFFD}, {0x100000, 0x10FFFD}
+  };
+  static const int max = sizeof(upright) / sizeof(upright[0]);
+  int st = 0;
+  int ed = max;
 
-  return g_unichar_iswide (ch);
+  while (st <= ed)
+    {
+      int mid = (st + ed) / 2;
+      if (upright[mid][0] <= ch && ch <= upright[mid][1])
+        return TRUE;
+      else
+        if (upright[mid][0] <= ch)
+          st = mid + 1;
+        else
+          ed = mid - 1;
+    }
+
+  return FALSE;
 }
 
 static void
@@ -883,7 +930,7 @@ width_iter_next(PangoWidthIter* iter)
   if (iter->end < iter->text_end)
     {
       gunichar ch = g_utf8_get_char (iter->end);
-      iter->wide = width_iter_iswide (ch);
+      iter->upright = width_iter_is_upright (ch);
     }
 
   while (iter->end < iter->text_end)
@@ -898,7 +945,7 @@ width_iter_next(PangoWidthIter* iter)
           continue;
         }
 
-      /* ignore the wide check if met joiner */
+      /* ignore the upright check if met joiner */
       if (met_joiner)
         {
           iter->end = g_utf8_next_char (iter->end);
@@ -915,7 +962,7 @@ width_iter_next(PangoWidthIter* iter)
           continue;
         }
 
-      if (width_iter_iswide (ch) != iter->wide)
+      if (width_iter_is_upright (ch) != iter->upright)
         break;
       iter->end = g_utf8_next_char (iter->end);
     }
@@ -1025,7 +1072,6 @@ itemize_state_init (ItemizeState      *state,
   state->gravity_hint = state->context->gravity_hint;
   state->resolved_gravity = PANGO_GRAVITY_AUTO;
   state->derived_lang = NULL;
-  state->lang_engine = NULL;
   state->current_fonts = NULL;
   state->cache = NULL;
   state->base_font = NULL;
@@ -1090,47 +1136,39 @@ copy_attr_slist (GSList *attr_slist)
 }
 
 static void
-itemize_state_fill_shaper (ItemizeState     *state,
-			   PangoEngineShape *shape_engine,
-			   PangoFont        *font)
+itemize_state_fill_font (ItemizeState *state,
+			 PangoFont    *font)
 {
   GList *l;
 
   for (l = state->result; l; l = l->next)
     {
       PangoItem *item = l->data;
-      if (item->analysis.shape_engine)
-	break;
+      if (item->analysis.font)
+        break;
       if (font)
 	item->analysis.font = g_object_ref (font);
-      else
-	item->analysis.font = NULL;
-      item->analysis.shape_engine = shape_engine;
     }
 }
 
 static void
-itemize_state_add_character (ItemizeState     *state,
-			     PangoEngineShape *shape_engine,
-			     PangoFont        *font,
-			     gboolean          force_break,
-			     const char       *pos)
+itemize_state_add_character (ItemizeState *state,
+			     PangoFont    *font,
+			     gboolean      force_break,
+			     const char   *pos)
 {
   if (state->item)
     {
-      if (!state->item->analysis.shape_engine && shape_engine)
+      if (!state->item->analysis.font && font)
 	{
-	  itemize_state_fill_shaper (state, shape_engine, font);
+	  itemize_state_fill_font (state, font);
 	}
-      else if (state->item->analysis.shape_engine && !shape_engine)
+      else if (state->item->analysis.font && !font)
 	{
 	  font = state->item->analysis.font;
-	  shape_engine = state->item->analysis.shape_engine;
 	}
 
       if (!force_break &&
-	  state->item->analysis.lang_engine == state->lang_engine &&
-	  state->item->analysis.shape_engine == shape_engine &&
 	  state->item->analysis.font == font)
 	{
 	  state->item->num_chars++;
@@ -1144,8 +1182,6 @@ itemize_state_add_character (ItemizeState     *state,
   state->item->offset = pos - state->text;
   state->item->length = 0;
   state->item->num_chars = 1;
-  state->item->analysis.shape_engine = shape_engine;
-  state->item->analysis.lang_engine = state->lang_engine;
 
   if (font)
     g_object_ref (font);
@@ -1215,15 +1251,11 @@ get_font_foreach (PangoFontset *fontset,
 		  gpointer      data)
 {
   GetFontInfo *info = data;
-  PangoEngineShape *engine;
-  PangoCoverageLevel level;
 
   if (G_UNLIKELY (!font))
     return FALSE;
 
-  engine = pango_font_find_shaper (font, info->lang, info->wc),
-  level = _pango_engine_shape_covers (engine, font, info->lang, info->wc);
-  if (level != PANGO_COVERAGE_NONE)
+  if (pango_font_has_char (font, info->wc))
     {
       info->font = font;
       return TRUE;
@@ -1248,36 +1280,17 @@ get_base_font (ItemizeState *state)
   return state->base_font;
 }
 
-static PangoScript
-get_script (ItemizeState      *state)
-{
-  /* Always use a basic shaper for vertical layout (ie, east/west gravity)
-   * as none of our script shapers support vertical shaping right now.
-   *
-   * XXX Should move the knowledge into the shaper interface.
-   */
-
-  if (PANGO_GRAVITY_IS_VERTICAL (state->resolved_gravity))
-    return PANGO_SCRIPT_COMMON;
-  else
-    return state->script;
-}
-
 static gboolean
-get_shaper_and_font (ItemizeState      *state,
-		     gunichar           wc,
-		     PangoEngineShape **shape_engine,
-		     PangoFont        **font)
+get_font (ItemizeState  *state,
+          gunichar       wc,
+          PangoFont    **font)
 {
   GetFontInfo info;
 
   /* We'd need a separate cache when fallback is disabled, but since lookup
    * with fallback disabled is faster anyways, we just skip caching */
   if (state->enable_fallback && font_cache_get (state->cache, wc, font))
-  {
-    *shape_engine = pango_font_find_shaper (*font, state->derived_lang, wc);
     return TRUE;
-  }
 
   info.lang = state->derived_lang;
   info.wc = wc;
@@ -1289,7 +1302,6 @@ get_shaper_and_font (ItemizeState      *state,
     get_font_foreach (NULL, get_base_font (state), &info);
 
   *font = info.font;
-  *shape_engine = pango_font_find_shaper (*font, state->derived_lang, wc);
 
   /* skip caching if fallback disabled (see above) */
   if (state->enable_fallback)
@@ -1348,7 +1360,7 @@ itemize_state_update_for_new_run (ItemizeState *state)
 	    gravity = state->context->resolved_gravity;
 
 	  state->resolved_gravity = pango_gravity_get_for_script_and_width (state->script,
-									    state->width_iter.wide,
+									    state->width_iter.upright,
 									    gravity,
 									    gravity_hint);
 	}
@@ -1368,11 +1380,6 @@ itemize_state_update_for_new_run (ItemizeState *state)
 	state->changed |= DERIVED_LANG_CHANGED;
     }
 
-  if ((state->changed & DERIVED_LANG_CHANGED) || !state->lang_engine)
-    {
-      state->lang_engine = _pango_get_language_engine ();
-    }
-
   if (state->changed & (EMOJI_CHANGED))
     {
       state->changed |= FONT_CHANGED;
@@ -1408,21 +1415,6 @@ itemize_state_update_for_new_run (ItemizeState *state)
     }
 }
 
-static const char *
-string_from_script (PangoScript script)
-{
-  static GEnumClass *class = NULL; /* MT-safe */
-  GEnumValue *value;
-  if (g_once_init_enter (&class))
-    g_once_init_leave(&class, (gpointer)g_type_class_ref (PANGO_TYPE_SCRIPT));
-
-  value = g_enum_get_value (class, script);
-  if (!value)
-    return string_from_script (PANGO_SCRIPT_INVALID_CODE);
-
-  return value->value_nick;
-}
-
 static void
 itemize_state_process_run (ItemizeState *state)
 {
@@ -1444,7 +1436,6 @@ itemize_state_process_run (ItemizeState *state)
     {
       gunichar wc = g_utf8_get_char (p);
       gboolean is_forced_break = (wc == '\t' || wc == LINE_SEPARATOR);
-      PangoEngineShape *shape_engine;
       PangoFont *font;
       GUnicodeType type;
 
@@ -1457,25 +1448,27 @@ itemize_state_process_run (ItemizeState *state)
        *
        * We don't want to change fonts just for variation selectors.
        * See bug #781123.
+       *
+       * Finally, don't change fonts for line or paragraph separators.
        */
       type = g_unichar_type (wc);
       if (G_UNLIKELY (type == G_UNICODE_CONTROL ||
                       type == G_UNICODE_FORMAT ||
                       type == G_UNICODE_SURROGATE ||
+                      type == G_UNICODE_LINE_SEPARATOR ||
+                      type == G_UNICODE_PARAGRAPH_SEPARATOR ||
                       (type == G_UNICODE_SPACE_SEPARATOR && wc != 0x1680u /* OGHAM SPACE MARK */) ||
                       (wc >= 0xfe00u && wc <= 0xfe0fu) ||
                       (wc >= 0xe0100u && wc <= 0xe01efu)))
         {
-	  shape_engine = NULL;
 	  font = NULL;
         }
       else
         {
-	  get_shaper_and_font (state, wc, &shape_engine, &font);
+	  get_font (state, wc, &font);
 	}
 
-      itemize_state_add_character (state,
-				   shape_engine, font,
+      itemize_state_add_character (state, font,
 				   is_forced_break || last_was_forced_break,
 				   p);
 
@@ -1484,34 +1477,28 @@ itemize_state_process_run (ItemizeState *state)
 
   /* Finish the final item from the current segment */
   state->item->length = (p - state->text) - state->item->offset;
-  if (!state->item->analysis.shape_engine)
+  if (!state->item->analysis.font)
     {
-      PangoEngineShape *shape_engine;
       PangoFont *font;
 
-      if (G_UNLIKELY (!get_shaper_and_font (state, ' ', &shape_engine, &font)))
-	{
-	  /* If no shaper was found, warn only once per fontmap/script pair */
-	  PangoFontMap *fontmap = state->context->font_map;
-	  const char *script_name = string_from_script (get_script (state));
+      if (G_UNLIKELY (!get_font (state, ' ', &font)))
+        {
+          /* If no font was found, warn once per fontmap/script pair */
+          PangoFontMap *fontmap = state->context->font_map;
+          const char *script_name = g_enum_to_string (G_TYPE_UNICODE_SCRIPT, state->script);
 
-	  if (!g_object_get_data (G_OBJECT (fontmap), script_name))
-	    {
-G_GNUC_BEGIN_IGNORE_DEPRECATIONS
-	      g_warning ("failed to choose a font, expect ugly output. engine-type='%s', script='%s'",
-			 pango_font_map_get_shape_engine_type (fontmap),
-			 script_name);
-G_GNUC_END_IGNORE_DEPRECATIONS
+          if (!g_object_get_data (G_OBJECT (fontmap), script_name))
+            {
+              g_warning ("failed to choose a font, expect ugly output. script='%s'",
+                         script_name);
 
-	      g_object_set_data_full (G_OBJECT (fontmap), script_name,
-				      GINT_TO_POINTER (1), NULL);
-	    }
+              g_object_set_data_full (G_OBJECT (fontmap), script_name,
+                                      GINT_TO_POINTER (1), NULL);
+            }
 
-	  shape_engine = _pango_get_fallback_shaper ();
-	  font = NULL;
-	}
-
-      itemize_state_fill_shaper (state, shape_engine, font);
+          font = NULL;
+        }
+      itemize_state_fill_font (state, font);
     }
   state->item = NULL;
 }
@@ -1616,7 +1603,7 @@ itemize_with_font (PangoContext               *context,
  * pango_itemize:
  * @context:   a structure holding information that affects
 	       the itemization process.
- * @text:      the text to itemize.
+ * @text:      the text to itemize. Must be valid UTF-8
  * @start_index: first byte in @text to process
  * @length:    the number of bytes (not characters) to process
  *             after @start_index.
@@ -1722,6 +1709,7 @@ update_metrics_from_items (PangoFontMetrics *metrics,
 	  /* metrics will already be initialized from the first font in the fontset */
 	  metrics->ascent = MAX (metrics->ascent, raw_metrics->ascent);
 	  metrics->descent = MAX (metrics->descent, raw_metrics->descent);
+	  metrics->height = MAX (metrics->height, raw_metrics->height);
 	  pango_font_metrics_unref (raw_metrics);
 	}
 
