@@ -46,10 +46,13 @@
 #include <math.h>
 
 #include "pango-context.h"
-#include "pangofc-fontmap.h"
+#include "pango-font-private.h"
+#include "pangofc-fontmap-private.h"
 #include "pangofc-private.h"
 #include "pango-impl-utils.h"
 #include "pango-enum-types.h"
+#include "pango-coverage-private.h"
+#include <hb-ft.h>
 
 
 /* Overview:
@@ -163,7 +166,8 @@ struct _PangoFcFontFaceData
   /* Data */
   FcPattern *pattern;	/* Referenced pattern that owns filename */
   PangoCoverage *coverage;
-  PangoFcCmapCache *cmap_cache;
+
+  hb_face_t *hb_face;
 };
 
 struct _PangoFcFace
@@ -259,7 +263,9 @@ static FcPattern       *pango_fc_patterns_get_font_pattern (PangoFcPatterns *pat
 static FcPattern *uniquify_pattern (PangoFcFontMap *fcfontmap,
 				    FcPattern      *pattern);
 
-static gpointer
+gpointer get_gravity_class (void);
+
+gpointer
 get_gravity_class (void)
 {
   static GEnumClass *class = NULL; /* MT-safe */
@@ -292,8 +298,7 @@ pango_fc_font_face_data_free (PangoFcFontFaceData *data)
   if (data->coverage)
     pango_coverage_unref (data->coverage);
 
-  if (data->cmap_cache)
-    _pango_fc_cmap_cache_unref (data->cmap_cache);
+  hb_face_destroy (data->hb_face);
 
   g_slice_free (PangoFcFontFaceData, data);
 }
@@ -1286,13 +1291,20 @@ is_alias_family (const char *family_name)
 {
   switch (family_name[0])
     {
+    case 'c':
+    case 'C':
+      return (g_ascii_strcasecmp (family_name, "cursive") == 0);
+    case 'f':
+    case 'F':
+      return (g_ascii_strcasecmp (family_name, "fantasy") == 0);
     case 'm':
     case 'M':
       return (g_ascii_strcasecmp (family_name, "monospace") == 0);
     case 's':
     case 'S':
       return (g_ascii_strcasecmp (family_name, "sans") == 0 ||
-	      g_ascii_strcasecmp (family_name, "serif") == 0);
+	      g_ascii_strcasecmp (family_name, "serif") == 0 ||
+	      g_ascii_strcasecmp (family_name, "system-ui") == 0);
     }
 
   return FALSE;
@@ -1334,7 +1346,7 @@ pango_fc_font_map_list_families (PangoFontMap      *fontmap,
       FcPatternDestroy (pat);
       FcObjectSetDestroy (os);
 
-      priv->families = g_new (PangoFcFamily *, fontset->nfont + 3); /* 3 standard aliases */
+      priv->families = g_new (PangoFcFamily *, fontset->nfont + 4); /* 4 standard aliases */
       temp_family_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
 
       count = 0;
@@ -1382,6 +1394,7 @@ pango_fc_font_map_list_families (PangoFontMap      *fontmap,
       priv->families[count++] = create_family (fcfontmap, "Sans", FC_PROPORTIONAL);
       priv->families[count++] = create_family (fcfontmap, "Serif", FC_PROPORTIONAL);
       priv->families[count++] = create_family (fcfontmap, "Monospace", FC_MONO);
+      priv->families[count++] = create_family (fcfontmap, "System-ui", FC_PROPORTIONAL);
 
       priv->n_families = count;
     }
@@ -1930,51 +1943,82 @@ pango_fc_font_map_get_font_face_data (PangoFcFontMap *fcfontmap,
   return data;
 }
 
-static PangoFcCmapCache *
-_pango_fc_cmap_cache_ref (PangoFcCmapCache *cmap_cache)
-{
-  g_atomic_int_inc ((int *) &cmap_cache->ref_count);
+typedef struct {
+  PangoCoverage parent_instance;
 
-  return cmap_cache;
+  FcCharSet *charset;
+} PangoFcCoverage;
+
+typedef struct {
+  PangoCoverageClass parent_class;
+} PangoFcCoverageClass;
+
+GType pango_fc_coverage_get_type (void) G_GNUC_CONST;
+
+G_DEFINE_TYPE (PangoFcCoverage, pango_fc_coverage, PANGO_TYPE_COVERAGE)
+
+static void
+pango_fc_coverage_init (PangoFcCoverage *coverage)
+{
 }
 
-void
-_pango_fc_cmap_cache_unref (PangoFcCmapCache *cmap_cache)
+static PangoCoverageLevel
+pango_fc_coverage_real_get (PangoCoverage *coverage,
+                            int            index)
 {
-  g_return_if_fail (cmap_cache->ref_count > 0);
+  PangoFcCoverage *fc_coverage = (PangoFcCoverage*)coverage;
 
-  if (g_atomic_int_dec_and_test ((int *) &cmap_cache->ref_count))
-    {
-      g_free (cmap_cache);
-    }
+  return FcCharSetHasChar (fc_coverage->charset, index)
+         ? PANGO_COVERAGE_EXACT
+         : PANGO_COVERAGE_NONE;
 }
 
-PangoFcCmapCache *
-_pango_fc_font_map_get_cmap_cache (PangoFcFontMap *fcfontmap,
-				   PangoFcFont    *fcfont)
+static void
+pango_fc_coverage_real_set (PangoCoverage *coverage,
+                            int            index,
+                            PangoCoverageLevel level)
 {
-  PangoFcFontFaceData *data;
+  PangoFcCoverage *fc_coverage = (PangoFcCoverage*)coverage;
 
-  if (G_UNLIKELY (fcfontmap == NULL))
-	return NULL;
+  if (level == PANGO_COVERAGE_NONE)
+    FcCharSetDelChar (fc_coverage->charset, index);
+  else
+    FcCharSetAddChar (fc_coverage->charset, index);
+}
 
-  if (G_UNLIKELY (!fcfont->font_pattern))
-    return NULL;
+static PangoCoverage *
+pango_fc_coverage_real_copy (PangoCoverage *coverage)
+{
+  PangoFcCoverage *fc_coverage = (PangoFcCoverage*)coverage;
+  PangoFcCoverage *copy;
 
-  data = pango_fc_font_map_get_font_face_data (fcfontmap, fcfont->font_pattern);
-  if (G_UNLIKELY (!data))
-    return NULL;
+  copy = g_object_new (pango_fc_coverage_get_type (), NULL);
+  copy->charset = FcCharSetCopy (fc_coverage->charset);
 
-  if (G_UNLIKELY (data->cmap_cache == NULL))
-    {
-      data->cmap_cache = g_new0 (PangoFcCmapCache, 1);
-      data->cmap_cache->ref_count = 1;
+  return (PangoCoverage *)copy;
+}
 
-      /* Make sure all cache entries are invalid initially */
-      data->cmap_cache->entries[0].ch = 1; /* char 1 cannot happen in bucket 0 */
-    }
+static void
+pango_fc_coverage_finalize (GObject *object)
+{
+  PangoFcCoverage *fc_coverage = (PangoFcCoverage*)object;
 
-  return _pango_fc_cmap_cache_ref (data->cmap_cache);
+  FcCharSetDestroy (fc_coverage->charset);
+
+  G_OBJECT_CLASS (pango_fc_coverage_parent_class)->finalize (object);
+}
+
+static void
+pango_fc_coverage_class_init (PangoFcCoverageClass *class)
+{
+  GObjectClass *object_class = G_OBJECT_CLASS (class);
+  PangoCoverageClass *coverage_class = PANGO_COVERAGE_CLASS (class);
+
+  object_class->finalize = pango_fc_coverage_finalize;
+
+  coverage_class->get = pango_fc_coverage_real_get;
+  coverage_class->set = pango_fc_coverage_real_set;
+  coverage_class->copy = pango_fc_coverage_real_copy;
 }
 
 PangoCoverage *
@@ -1984,9 +2028,6 @@ _pango_fc_font_map_get_coverage (PangoFcFontMap *fcfontmap,
   PangoFcFontFaceData *data;
   FcCharSet *charset;
 
-  if (G_UNLIKELY (!fcfont->font_pattern))
-    return NULL;
-
   data = pango_fc_font_map_get_font_face_data (fcfontmap, fcfont->font_pattern);
   if (G_UNLIKELY (!data))
     return NULL;
@@ -1998,7 +2039,7 @@ _pango_fc_font_map_get_coverage (PangoFcFontMap *fcfontmap,
        * doesn't require loading the font
        */
       if (FcPatternGetCharSet (fcfont->font_pattern, FC_CHARSET, 0, &charset) != FcResultMatch)
-	return NULL;
+        return NULL;
 
       data->coverage = _pango_fc_font_map_fc_to_coverage (charset);
     }
@@ -2018,50 +2059,12 @@ _pango_fc_font_map_get_coverage (PangoFcFontMap *fcfontmap,
 PangoCoverage  *
 _pango_fc_font_map_fc_to_coverage (FcCharSet *charset)
 {
-  PangoCoverage *coverage;
-  FcChar32  ucs4, pos;
-  FcChar32  map[FC_CHARSET_MAP_SIZE];
-  int i;
+  PangoFcCoverage *coverage;
 
-  /*
-   * Convert an Fc CharSet into a pango coverage structure.  Sure
-   * would be nice to just use the Fc structure in place...
-   */
-  coverage = pango_coverage_new ();
-  for (ucs4 = FcCharSetFirstPage (charset, map, &pos);
-       ucs4 != FC_CHARSET_DONE;
-       ucs4 = FcCharSetNextPage (charset, map, &pos))
-    {
-      for (i = 0; i < FC_CHARSET_MAP_SIZE; i++)
-	{
-	  FcChar32  bits = map[i];
-	  FcChar32  base = ucs4 + i * 32;
-	  int b = 0;
+  coverage = g_object_new (pango_fc_coverage_get_type (), NULL);
+  coverage->charset = FcCharSetCopy (charset);
 
-	  while (bits)
-	    {
-	      if (bits & 1)
-		pango_coverage_set (coverage, base + b, PANGO_COVERAGE_EXACT);
-
-	      bits >>= 1;
-	      b++;
-	    }
-	}
-    }
-
-  /* Awful hack so Hangul Tone marks get rendered with the same
-   * font and in the same run as other Hangul characters. If a font
-   * covers the first composed Hangul glyph, then it is declared to cover
-   * the Hangul tone marks. This hack probably needs to be formalized
-   * by choosing fonts for scripts rather than individual code points.
-   */
-  if (pango_coverage_get (coverage, 0xac00) == PANGO_COVERAGE_EXACT)
-    {
-      pango_coverage_set (coverage, 0x302e, PANGO_COVERAGE_EXACT);
-      pango_coverage_set (coverage, 0x302f, PANGO_COVERAGE_EXACT);
-    }
-
-  return coverage;
+  return (PangoCoverage *)coverage;
 }
 
 /**
@@ -2689,3 +2692,23 @@ pango_fc_family_init (PangoFcFamily *fcfamily)
 {
   fcfamily->n_faces = -1;
 }
+
+hb_face_t *
+pango_fc_font_map_get_hb_face (PangoFcFontMap *fcfontmap,
+                               PangoFcFont    *fcfont)
+{
+  PangoFcFontFaceData *data;
+
+  data = pango_fc_font_map_get_font_face_data (fcfontmap, fcfont->font_pattern);
+  
+  if (!data->hb_face)
+    {
+      hb_blob_t *blob;
+
+      blob = hb_blob_create_from_file (data->filename);
+      data->hb_face = hb_face_create (blob, data->id);
+      hb_blob_destroy (blob);
+    }
+
+  return data->hb_face;
+}
